/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * cnx_lst.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <glib.h>

#include "gvar.h"
#include "cnx_lst.h"
#include "manage_cnx.h"
#include "timed_out_string.h"
#include "tos_key.h"
#include "hub_cluster.h"
#include "multi_public_chat.h"

/**********************************/
/* list of all running connexions */
/**********************************/
cnx_hash_table *cnx_lst=NULL;
HL_MUTEX cnx_struct_lock = HL_MUTEX_INIT;

void *g_string_ref_free(GString_ref *str)
{
	if (str)
	{
		if (--(str->count) <= 0)
		{
			g_string_free(str->data, TRUE);
			free(str);
		}
	}
	return(0);
}

GString_ref	*conv_gstring_to_gstring_ref(GString *str)
{
	GString_ref	*out;

	if (!str)
		return (0);
	out = malloc(sizeof(GString_ref));
	out->count = 1;
	out->data = str;
	return(out);
}

GString_ref	*g_string_ref_new(const char *str)
{
	GString_ref	*out;

	if (!str)
		return (0);
	out = malloc(sizeof(GString_ref));
	out->count = 1;
	out->data = g_string_new(str);
	return(out);
}

/***********************************/
/*	return the hash value of char * */
/********************************************/
/* output: value between 0 and HASH_SIZE - 1*/
/********************************************/
static int hach_str(const char *str)
{
	int len=0;
	int hash=0;

	len = strlen(str);
	while(len--)
	{
		register unsigned char c;

		c = str[len];
		if ((c >= 'A') && (c <= 'Z'))
			hash += c - ('A' - 'a');
		else
			hash += c;
	}
	return (hash % HASH_SIZE);
}

/******************************************************/
/*	return the hash value of the user_nick of an entry	*/
/******************************************************/
/* output: value between 0 and HASH_SIZE - 1*/
/********************************************/
static int hach(const CNX_ENTRY *ent)
{
	return (hach_str(ent->user_nick->str));
}

/*******************************************/
/* create a new empty hash table structure */
/*******************************************/
/* this function should be call only once	*/
/* the ouput must be affect to cnx_lst		 */
/*******************************************/
cnx_hash_table	*new_cnx_hash_table(void)
{
	cnx_hash_table	*out;

	out = malloc(sizeof(cnx_hash_table)); /* should not fail */
	out->wait_len = 0;
	out->len = 0;
	out->locallen = 0;
	out->wait = 0;
	out->entry = malloc(HASH_SIZE * sizeof(void *));
	bzero(out->entry, HASH_SIZE * sizeof(void *));
	return(out);
}

/***********************************************/
/* add a CNX_ENTRY in the hash table structure */
/******************************************************************/
/* this function must be called only when cnx_lst_lock is already */
/* locked for write.																*/
/******************************************************************/
int add_in_cnx_hash_table(cnx_hash_table *table, CNX_ENTRY *new)
{
	int i;

	table->len++;
	if (new->server_id==0)
		table->locallen++;
	i = hach(new);
	new->next = table->entry[i];
	table->entry[i] = new;
	return (0);
}

/***************************************************/
/* get the CNX_ENTRY from the hash table structure */
/* with the asked usernick									*/
/******************************************************************/
/* this function must be called only when cnx_lst_lock is already */
/* locked for read.															   */
/******************************************************************/
CNX_ENTRY		 *get_in_cnx_hash_table_by_nick(const cnx_hash_table *table, const char *str)
{
	CNX_ENTRY	 *ptr;
	int		i;

	i = (hach_str(str));
	ptr = table->entry[i];
	while (ptr)
		{
			if (strcasecmp(str, ptr->user_nick->str))
		ptr = ptr->next;
			else
		return (ptr);
		}
	return (0);
}

/******************************************************************/
/* remove and destroy the CNX_ENTRY from the hash table structure */
/* with the asked usernick														*/
/******************************************************************/
/* this function must be called only when cnx_lst_lock is already */
/* locked for write.																*/
/* output: 0=operation succes else 1=user not found					*/
/******************************************************************/
int del_in_cnx_hash_table_by_nick(cnx_hash_table *table, const char *str)
{
	CNX_ENTRY	*tmp;
	CNX_ENTRY	*old;
	int		i;

#ifdef __COLOR_DEBUG_H__
	printf("del_in_cnx_hash_table_by_nick ("COLOYEL1"%s"COLONONE")\n", str);
#endif
	i = (hach_str(str));
	tmp = table->entry[i];
	old = 0;
	while (tmp)
	{
		if (strcasecmp(str, tmp->user_nick->str))
			tmp = tmp->next;
		else
		{
			if (old == 0)
				table->entry[i] = table->entry[i]->next; 
			else
			{
				old->next = tmp->next;
			}
			if (tmp->server_id==0)
				table->locallen--;
			delete_cnx_unlinked_entry(tmp);
			table->len--;
			return(0);
		}
	}

#ifdef __COLOR_DEBUG_H__
	printf("del_in_cnx_hash_table_by_nick! %s [" COLORED1 "FAIL" COLONONE"]\n", str);
#endif
	return (1);
}

/******************************************************/
/* unlink the CNX_ENTRY from the hash table structure	*/
/* with the asked usernick										*/
/******************************************************************/
/* this function must be called only when cnx_lst_lock is already */
/* locked for write.																*/
/******************************************************************/
int unlink_from_cnx_hash_table(cnx_hash_table *table, CNX_ENTRY *ptr)
{
	CNX_ENTRY	*ent;
	CNX_ENTRY	*old;
	int i;

#ifdef __COLOR_DEBUG_H__
	printf("unlink_from_cnx_hash_table - ("COLOYEL1"%s"COLONONE")\n", ptr->user_nick->str);
#endif

	i = hach(ptr);
	for(old=0,ent=table->entry[i];ent;old=ent,ent=ent->next)
	{
		if(ent == ptr)
		{
			if (old == NULL)
			{
				table->entry[i] = ent->next;
			}
			else
			{
				old->next = ent->next;
			}
			if (ptr->server_id==0)
			{
				table->locallen--;
			}
			table->len--;
			return(TRUE);
		}
	}
	return (FALSE);
}

int add_in_cnx_wait_table(cnx_hash_table *table, CNX_ENTRY *new)
{
	if (new->is_op & HUB_PRIV)
		printf("a new HUB in QUEUE fd :%d\n", new->sock_fd);
	table->wait_len++;
	new->next = table->wait;
	table->wait = new;
	return (0);
}

/************************************************************************/
/* remove and destroy the given entry from the cnx_lst or the wait list */
/************************************************************************/
/* input: address of the entry to remove */
/* output: 0=ok, 1=error					  */
/*****************************************/
int remove_from_all_cnx_table(cnx_hash_table *table, CNX_ENTRY *ent)
{
	CNX_ENTRY *prev = NULL;
	CNX_ENTRY *ptr = NULL;

	if (table->wait == 0)
	{
		return (delete_cnx_entry(ent));
	}
	if (ent == table->wait)
	{
		table->wait = ent->next;
		table->wait_len--;
		delete_cnx_unlinked_entry(ent);

		return (0);
	}
	ptr = table->wait;
	while (ptr && ptr != ent)
	{
		prev = ptr;
		ptr = ptr->next;
	}
	if (ptr)
	{
		if (prev)
			prev->next = ent->next;
		else
			table->wait = ent->next;
		table->wait_len--;
		delete_cnx_unlinked_entry(ent);
		return (0);
	}
	printf ("remove_from_all_cnx_table (somethink wong ...)\n");
	return (delete_cnx_entry(ent));
}

/*************************************************************************/
/* inject the handshaked user form que queue to the hash table strucutre */
/* inform the others servers of the arival of the new user					 */
/*************************************************************************/
/* output the number of injected user(s)				*/
/* should almost alway return 1							*/
/***************************************************/
int inject_wait_table(cnx_hash_table *table)
{
	CNX_ENTRY *ptr=NULL;
	CNX_ENTRY *old_ptr=NULL;
	int ret=0;

	old_ptr = 0;
	for (ptr=table->wait; ptr; old_ptr = ptr, ptr = ptr->next)
	{
		if (ptr->handshake_done!=1)
			continue;
		if (old_ptr == NULL)
		{
			table->wait = ptr->next;
		}
		else
		{
			old_ptr->next = ptr->next;
		}
		table->wait_len--;
		ptr->next = NULL;
			
		if (ptr->is_op & HUB_PRIV) /* external hub */
		{
			int	tmp;

			ptr->connected=1;
			memcpy(route_table->table + ptr->server_id, ptr, sizeof (CNX_ENTRY));
			tmp = ptr->server_id;
			route_table->server_count++;
			ptr->sock_fd = -1;
			free(ptr);
			return(1);
		}
		add_in_cnx_hash_table(table, ptr);
			
		/* Cluster Part 2 ADD the user to the other server */
		if (route_table->server_count > 1)
		{
			int i;
			GString *paked_str;
			GString_ref *paked_str2;
#ifdef __CYGWIN__
			char		long_int[48];
#endif		
	
			paked_str = g_string_new("");
#ifdef __CYGWIN__
			sprintf(long_int, "%Lu", ptr->shared_size);
			g_string_sprintf(paked_str,"new:%s$%s$%s$%s$%c$%s$%s$%d$%s$|",
				 					ptr->user_nick->str,				/*1 s	*/
				 					ptr->user_description->str,	/*2 s	*/
				 					ptr->user_mail->str,				/*3 s	*/
				 					ptr->user_cnx_type,				/*4 s	*/
				 					(char)ptr->user_flag,			/*5 c	*/
				 					long_int,							/*6 Lu */
				 					inet_ntoa(ptr->ip),				/*7 s	*/
				 					ptr->is_op,							/*8 d	*/
				 					ptr->client_version->str);		/*9 s */
#else
			g_string_sprintf(paked_str,"new:%s$%s$%s$%s$%c$%Lu$%s$%d$%s$|",
				 					ptr->user_nick->str,		/*1 s	*/
				 					ptr->user_description->str,	/*2 s	*/
				 					ptr->user_mail->str,		/*3 s	*/
				 					ptr->user_cnx_type,		/*4 s	*/
				 					(char)ptr->user_flag,		/*5 c	*/
				 					ptr->shared_size,			/*6 Lu */
				 					inet_ntoa(ptr->ip),		/*7 s	*/
				 					ptr->is_op,			/*8 d	*/
				 					ptr->client_version->str);		/*9 s */
#endif
			paked_str2=conv_gstring_to_gstring_ref(paked_str);
			for (i = 1; i <= route_table->max_server_id; i++)
				if (route_table->table[i].connected)
			send_str_to((route_table->table + i),paked_str2);
			g_string_ref_free(paked_str2);
			return(1);
		}
		return(1);	
	}
	return (ret);
}

void free_cmd_glist(GList **pa);
void free_cmd_glist_ref(GList **pa);

/***************************************************/
/* create a copy of the given CNX_ENTRY and put it */
/* inside the cnx_lst.                             */
/******************************************************/
/* input: orig= entry to add the cnx_lst              */
/*        no_dup= 0, create a copy of orig and put the*/
/*                   copy in the cnx_lst.             */
/*              = 1, put orig itself in cnx_lst, don't*/
/*                   create copy.                     */
/* output: address of the newly created entry or NULL */
/******************************************************/
CNX_ENTRY *add_new_cnx_entry(CNX_ENTRY *orig, int no_dup)
{
	CNX_ENTRY *nw;

	if(orig==NULL)
		return NULL;

	/* create the copy of orig */
	if(no_dup==0)
	{
		nw=malloc(sizeof(CNX_ENTRY));
		if(nw==NULL)
			return NULL;
		memcpy(nw,orig,sizeof(CNX_ENTRY));
	}
	else
		nw=orig;

	/* and put it inside the list */
	/*if(cnx_lst==NULL)  cnx_lst=new_cnx_hash_table();*/ /*	if (nw->user_nick->str[0] != 0)	
	  add_in_cnx_hash_table(cnx_lst,nw); / / should never enter there	else / / toujours ici
	*/

	add_in_cnx_wait_table(cnx_lst, nw);

	return nw;
}

/*********************************************************************/
/* delete the given entry from the cnx_lst and free allocated memory */
/*********************************************************************/
/* input: address of the entry to remove */
/* output: 0=ok, 1=error                 */
/*****************************************/
int delete_cnx_entry(CNX_ENTRY *ent)
{
	int ret=1;

	if(ent==NULL)
		return ret;

#ifdef __COLOR_DEBUG_H__
	printf(COLOYEL1 "delete_cnx_entry %d" COLONONE"\n", ent->sock_fd);
#endif

	if (unlink_from_cnx_hash_table(cnx_lst, ent)==TRUE)
	{
#ifdef __COLOR_DEBUG_H__
		if (ent->user_nick)
			printf(COLOYEL1 "delete_cnx_entry unlink %s %d" COLONONE"\n", ent->user_nick->str, ent->sock_fd);
		else
			printf(COLOYEL1 "delete_cnx_entry unlink %s %d" COLONONE"\n", "ERRORRRR", ent->sock_fd);
#endif

		if(ent->sock_fd!=-1)
		{
			if (ent->sock_fd != 0)
				close(ent->sock_fd);
#ifdef __COLOR_DEBUG_H__
			else
				printf(COLOYEL1"SAMERE"COLONONE"\n");
#endif
			ent->sock_fd=-1;
		}
		return (delete_cnx_unlinked_entry(ent));		
	}
	if (ent->is_op & HUB_PRIV)
	{
		lost_hub(ent);
		route_table->server_count--;
	}
	return ret;
}


/******************************************************************/
/*	 use delete_cnx_unlinked_entry when the cnx isen't linked from */
/*	 the hach table or que queue												*/
/******************************************************************/
/* input: address of the entry to remove */
/* output: 0=ok, 1=error					  */
/*****************************************/
int delete_cnx_unlinked_entry(CNX_ENTRY *ent)
{
	int ret=1;

	if(ent==NULL)
		return ret;

	pubchat_user_quit(ent);/* remove the user frome the chan */
	if(ent->sock_fd!=-1)
	{
		close(ent->sock_fd);
		ent->sock_fd=-1;
	}

	if(ent->user_nick!=NULL)
		g_string_free(ent->user_nick,TRUE);
	if(ent->user_mail!=NULL)
		g_string_free(ent->user_mail,TRUE);
	if(ent->user_description!=NULL)
		g_string_free(ent->user_description,TRUE);
	if(ent->client_version!=NULL)
		g_string_free(ent->client_version,TRUE);

	if(ent->incoming_commands!=NULL)
		free_cmd_glist(&(ent->incoming_commands));
	if(ent->outgoing_commands!=NULL)
		free_cmd_glist_ref(&(ent->outgoing_commands));
	if(ent->in_partial!=NULL)
	{
		g_string_free(ent->in_partial,TRUE);
		ent->in_partial=NULL;
	}
	if(ent->created_lock!=NULL)
		free(ent->created_lock);
	if(ent->temp_nick!=NULL)
		free(ent->temp_nick);
	if(ent->user_pass!=NULL)
		free(ent->user_pass);
	free(ent);
	ret=0;

	return ret;
}


/**********************************************************************/
/* the following functions deal with GList of *_commands of CNX_ENTRY */
/**********************************************************************/

/************************************************************************/
/* append a copy of the given string to the given glist for outgoing msg*/
/************************************************************************/
void append_new_cmd_to_glist_outgoing(CNX_ENTRY *ent,GList **pa, GString_ref *string)
{
	string->count++;
	(*pa)=g_list_append((*pa),string);
}

/*************************************************************************/
/* append a copy of the given string to the given glist for incomming msg*/
/*************************************************************************/

void append_new_cmd_to_glist_incoming(CNX_ENTRY *ent,GList **pa, char *string)
{
	GString *s;

	s=g_string_new(string);
	(*pa)=g_list_append((*pa),s);
}


/*********************************************************/
/* prepend a copy of the given string to the given glist */
/*********************************************************/
void prepend_new_cmd_to_glist(CNX_ENTRY *ent, GList **pa, char *string)
{
	GString *s;

	s=g_string_new(string);
	
	(*pa)=g_list_prepend((*pa),s);
}

/**************************************************************/
/* return the pointer of the first GString of the given GList */
/**************************************************************/
/* output: the pointer or NULL. the pointer and the GString */
/*				 content should not be modified						*/
/************************************************************/
const GString_ref * const peek_first_string_of_glist_outgoing(CNX_ENTRY *ent, GList **pa)
{
	const GString_ref *st;

	if((*pa)==NULL)
		st=NULL;
	else
		st=(*pa)->data;
	return st;
}


const GString * const peek_first_string_of_glist_incoming(CNX_ENTRY *ent, GList **pa)
{
	const GString *st;

	if((*pa)==NULL)
		st=NULL;
	else
		st=(*pa)->data;
	return st;
}



/************************************************************/
/* same as peek_first_string_of_glist except the GString is */
/* also removed from the GList. You must free it yourself	*/
/* when you don't need it anymore.									*/
/************************************************************/
/* output: the pointer or NULL */
/*******************************/
GString_ref *take_first_string_of_glist_outgoing(CNX_ENTRY *ent, GList **pa)
{
	GString_ref *st;

	if((*pa)==NULL)
		st=NULL;
	else
	{
		st=(*pa)->data;
		(*pa)=g_list_remove((*pa),st);
	}
	return st;
}


GString *take_first_string_of_glist_incoming(CNX_ENTRY *ent, GList **pa)
{
	GString *st;

	if((*pa)==NULL)
		st=NULL;
	else
	{
		st=(*pa)->data;
		(*pa)=g_list_remove((*pa),st);
	}
	return st;
}


/********************************************************************/
/* free all the commands (string) stored inside the given cmd_glist */
/********************************************************************/
/* this function must be called only when cnx_lst_lock is already */
/* locked for write.																*/
/******************************************************************/
void free_cmd_glist(GList **pa)
{
	if(pa==NULL)
		return;

	while((*pa)!=NULL)
	{
		gpointer cmd;

		cmd=(*pa)->data;

		(*pa)=g_list_remove((*pa),cmd);

		g_string_free(cmd,TRUE);
	}
}

void free_cmd_glist_ref(GList **pa)
{
	if(pa==NULL)
		return;

	while((*pa)!=NULL)
	{
		gpointer cmd;

		cmd=(*pa)->data;

		(*pa)=g_list_remove((*pa),cmd);

		g_string_ref_free(cmd);
	}
}

/************************************************/
/* call send_str_to for all existing connexions */
/************************************************/
void queue_to_everyone(GString_ref *str)
{	
	int i;
	CNX_ENTRY *ent;

	for(i=0;i<HASH_SIZE;i++) /* for queue_to_everyone */
	{
		for(ent=cnx_lst->entry[i];ent;ent=ent->next)
		{
			if((ent->sock_fd!=-1)&&(ent->server_id==0))
			{
				send_str_to(ent,str);
			}
		}
	}

	for(ent=cnx_lst->wait; ent; ent=ent->next) 
	{
		if((ent->sock_fd!=-1)&&(ent->user_nick!=NULL)&&(ent->user_nick->len!=0)) // MINI FIXME
			send_str_to(ent,str);
	}

	/* NEW4 PART brd*/
	/* send str to serveur for broadcast */

	if (route_table->server_count > 1)
	{
		GString	*paked_str;
		GString_ref *paked_str2;
			
		paked_str = g_string_new("");
		g_string_sprintf(paked_str, "brd:%s", str->data->str);
		paked_str2 = conv_gstring_to_gstring_ref(paked_str);
		for (i = 1; i <= route_table->max_server_id; i++)
		{
			if (route_table->table[i].connected)
				send_str_to((route_table->table + i),paked_str2);
		}
		g_string_ref_free(paked_str2);
	/* implemente send_str_to(ent,str); */
	}
	/* NEW4 PART */
}

void queue_to_everyone_exetp(GString_ref *str, const char *expt)
{
	int i;
		CNX_ENTRY *ent;
	CNX_ENTRY *expt_ent;

	expt_ent = get_in_cnx_hash_table_by_nick(cnx_lst, expt);
	for(i=0;i<HASH_SIZE;i++) /* for queue_to_everyone */
	{
		for(ent=cnx_lst->entry[i];ent;ent=ent->next)
		{
			if((ent->sock_fd!=-1)&&(ent->server_id==0))
			{
				if (expt_ent != ent)
					send_str_to(ent,str);
			}
		}
	}

	/* NEW4 PART brd*/
	/* send str to serveur for broadcast */

	if (route_table->server_count > 1)
	{
		GString	*paked_str;
		GString_ref *paked_str2;
			
		paked_str = g_string_new("");
		g_string_sprintf(paked_str, "brd:%s", str->data->str);
		paked_str2 = conv_gstring_to_gstring_ref(paked_str);
		for (i = 1; i <= route_table->max_server_id; i++)
		{
			if (route_table->table[i].connected)
				send_str_to((route_table->table + i),paked_str2);
		}
		g_string_ref_free(paked_str2);
		/* implemente send_str_to(ent,str); */
	}
	/* NEW4 PART */
}


void queue_to_extern_everyone(GString_ref *str, char* from)
{
	int i;
	if (route_table->server_count > 1)
	{
		char *paked_str;
		GString_ref *paked_str2;

		paked_str = malloc(strlen(str->data->str) + 6 + strlen(from));
		sprintf(paked_str, "br2:%s$%s", from, str->data->str);
		paked_str2 = g_string_ref_new(paked_str);
		for (i = 1; i <= route_table->max_server_id; i++)
		{
			if (route_table->table[i].connected)
				send_str_to((route_table->table + i),paked_str2);
		}
		g_string_ref_free(paked_str2);
	}
}


void queue_to_local_everyone(GString_ref *str)
{	
	int i;
	CNX_ENTRY *ent;


	for(i=0;i<HASH_SIZE;i++) /* for queue_to_everyone */
	{
		for(ent=cnx_lst->entry[i];ent;ent=ent->next)
		{
			if((ent->sock_fd!=-1)&&(ent->server_id==0))
			{
				send_str_to(ent,str);
			}
		}
	}

	for(ent=cnx_lst->wait; ent; ent=ent->next) 
	{
		if((ent->sock_fd!=-1)&&(ent->user_nick!=NULL)&&(ent->user_nick->len!=0)) // MINI FIXME
			send_str_to(ent,str);
	}
}


/****************************************************************/
/* call send_str_to for the connexion having the given nickname */
/****************************************************************/
CNX_ENTRY *send_str_to_named_user(const char *dest, char *string)
{
	CNX_ENTRY *ent;

	ent=get_in_cnx_hash_table_by_nick(cnx_lst, dest);
	if(ent!=NULL)
	{
		GString_ref	*str;
		str = g_string_ref_new(string);
		if(ent->server_id==0)
			send_str_to(ent,str);
		else
		{
			/* NEW5 PART pvm*/
			/* send str to serveur for PM */
			if (route_table->table[ent->server_id].connected)
			{
				GString	*paked_str;
				GString_ref *paked_str2;

				paked_str = g_string_new("");
				g_string_sprintf(paked_str,"pvm:%s$%s", dest, string);
				paked_str2 = conv_gstring_to_gstring_ref(paked_str);
				send_str_to((route_table->table + ent->server_id), paked_str2);
				g_string_ref_free(paked_str2);
			}
			/* NEW5 PART */
		}
		g_string_ref_free(str);
	}
	return (ent);
}

/***********************************************************************/
/* return the connetion having the given nickname or NULL if not exist */
/***********************************************************************/
/* WARNING: before using this function, you must lock cnx_lst */
/*					before and release the lock when no more useful.*/
/**************************************************************/
CNX_ENTRY *find_cnx_by_nickname(const char *nick)
{
	return(get_in_cnx_hash_table_by_nick(cnx_lst, nick));
}

/********************************************************************/
/* retrieve the list of all connected users								  */
/* The list is returned as a string, each nickname separated by a $ */
/********************************************************************/
GString *get_userlist(void)
{
	GString *str;
	int i,j;
	CNX_ENTRY *ent;

	j=0;
	str=g_string_new("");

	/* find all users and put them in the string */
	for(i=0;i<HASH_SIZE;i++)
	{
		for(ent=cnx_lst->entry[i];ent;ent=ent->next)	
		{
			if((ent->sock_fd!=-1)&&
				(ent->user_nick!=NULL)&&
				(ent->user_nick->len!=0))
			{
				j++;
				if(str->len!=0)
				{
					str=g_string_append_c(str,'$');
				}
				str=g_string_append(str,ent->user_nick->str);
			}
		}
	}
	printf ("%d users found\n",j);
	return str;
}

/********************************************************************/
/* retrieve the list of all connected ops									  */
/* The list is returned as a string, each nickname separated by a $ */
/********************************************************************/
GString *get_oplist(void)
{
	GString *str;
	int i;
	CNX_ENTRY *ent;

	str=g_string_new("");

	/* find all users and put them in the string */
	for(i=0;i<HASH_SIZE;i++)
		for(ent=cnx_lst->entry[i];ent;ent=ent->next)	
	{
		if((ent->sock_fd!=-1)&&
			(ent->user_nick!=NULL)&&
			(ent->user_nick->len!=0)&&
			(ent->is_op))
		{
			if(str->len!=0)
			{
				str=g_string_append_c(str,'$');
			}
			str=g_string_append(str,ent->user_nick->str);
		}
	}


	return str;
}


/***************************************************/
/* get the user information for the given nickname */
/***************************************************/
/* output: NULL if user not found */
/**********************************/
GString *get_user_info(char *nick)
{
	CNX_ENTRY *ent;
	GString *str=NULL;
#ifdef __CYGWIN__
			char		long_int[48];
#endif

	ent=find_cnx_by_nickname(nick);
	if(ent!=NULL)
	{
		str=g_string_new("");
#ifdef __CYGWIN__
		sprintf(long_int, "%Lu", ent->shared_size);
		g_string_sprintf(str,"%s$%s$%s$%s$%u$%d$%s",
					((ent->user_cnx_type!=NULL)?ent->user_cnx_type:""),
					long_int,
					((ent->user_mail->str!=NULL)?ent->user_mail->str:""),
					((ent->user_description->str!=NULL)?ent->user_description->str:""),
					ent->user_flag,
					ent->is_op,
					((ent->client_version->str!=NULL)?ent->client_version->str:"")
					);
#else
		g_string_sprintf(str,"%s$%Lu$%s$%s$%u$%d$%s",
					((ent->user_cnx_type!=NULL)?ent->user_cnx_type:""),
					ent->shared_size,
					((ent->user_mail->str!=NULL)?ent->user_mail->str:""),
					((ent->user_description->str!=NULL)?ent->user_description->str:""),
					ent->user_flag,
					ent->is_op,
					((ent->client_version->str!=NULL)?ent->client_version->str:"")
					);
#endif
	}
	return str;
}

/******************************************************/
/* get the connection duration for the given nickname */
/**************************************************************/
/* output: 0= not found, else 1 and *duration is the duration */
/**************************************************************/
int get_user_connection_duration(char *nick, time_t *duration)
{
	CNX_ENTRY *ent;
	int ret=0;

	ent=find_cnx_by_nickname(nick);
	if(ent!=NULL)
	{
		ret=1;
		(*duration)=time(NULL)-ent->start_time;
	}
	return ret;
}

/***************************/
/* get hub user statistics */
/***************************/
void get_hub_user_stat(unsigned long long *shared_size, unsigned int *nb_users) // BOOO
{
	int i;
	CNX_ENTRY *ent;

	if(cnx_lst==NULL)
	{
		*nb_users=0;
		*shared_size=0;
		return;
	}

	*nb_users=cnx_lst->len;
	*shared_size=0;

	for(i=0;i<HASH_SIZE;i++)
	for(ent=cnx_lst->entry[i];ent;ent=ent->next)
	{
		if((ent->sock_fd!=-1)&&
			(ent->user_nick!=NULL)&&
			(ent->user_nick->len!=0))
		{
			(*shared_size)+=ent->shared_size;
		}
	}
}

/*************************************/
/* get the number of connected users */
/*************************************/
int get_nb_connected(void)
{
	int nb=0;

	if(cnx_lst!=NULL)
		nb=cnx_lst->len;
	return nb;
}

/*******************************************/
/* get the number of local connected users */
/*******************************************/
int get_local_nb_connected(void)
{
	int nb=0;

	if(cnx_lst!=NULL)
		nb=cnx_lst->locallen;
	return nb;
}

/*********************************/
/* disconnect the given nickname */
/*********************************/
void do_disconnect(char *nickname)
{
	CNX_ENTRY *cnx;

	cnx=find_cnx_by_nickname(nickname);
	if(cnx!=NULL)
	{
		if(cnx->sock_fd!=-1)
		{
			cnx_io_processing(cnx,1,1);		/* flush queued message but only if possible */
			shutdown(cnx->sock_fd,0);		/* we don't want to receive more data from this connexion */
		}
		else
		{
			if (cnx->server_id != 0)
				send_kick_entry_to_hub(cnx);
		}
	}
}

/***************************/
/* kick the given nickname */
/**************************************************************/
/* this function closes incoming side of the nickname socket  */
/* and broadcast the Hub-Security message.                    */
/**************************************************************/
void kick_named_user(CNX_ENTRY *ptr,char *nickname)
{
	CNX_ENTRY *cnx;
	int temp_ban;

	/* retrieve the temp ban duration when a kick occurs */
	G_LOCK(conf_file);
	if(!e_db_int_get(conf_file,"KICKTEMPBAN",&temp_ban))
		temp_ban=0;
	G_UNLOCK(conf_file);

	cnx=find_cnx_by_nickname(nickname);
	if(cnx!=NULL)
	{
		GString *str;
		GString_ref *str2;

		if ((cnx->sock_fd!=-1) || (cnx->server_id != 0))
		{
			if (temp_ban != 0)
			{
				send_ban_entry_to_hub(cnx,temp_ban*60);
				add_tos_entry_v1_uniq(TBAN_TOSKEY,temp_ban*60,(void*)(&(cnx->ip)),sizeof(struct in_addr),NULL,0);
			}
			if (cnx->server_id == 0)
			{
				cnx_io_processing(cnx,1,1);		/* flush queued message but only if possible */
				shutdown(cnx->sock_fd,0);		/* we don't want to receive more data from this connexion */
			}
			else
			{
				send_kick_entry_to_hub(cnx);
			}
			str=g_string_new("");
			if(cnx!=NULL)
			{
				g_string_sprintf(str,"<Hub-Security> The user %s was kicked by %s. IP: %s|",
					 nickname,ptr->user_nick->str,inet_ntoa(cnx->ip));
			}
			else
			{
				g_string_sprintf(str,"<Hub-Security> The user %s was kicked. IP: %s|",
					 nickname,inet_ntoa(cnx->ip));
			}
			str2 = conv_gstring_to_gstring_ref(str);
			queue_to_everyone(str2);
			g_string_ref_free(str2);
		}
	}
}


