/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * manage_cnx.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <ctype.h>
#include <glib.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "cnx_lst.h"
#include "manage_cnx.h"
#include "main.h"
#include "gvar.h"
#include "key.h"
#include "macro.h"
#include "passwd.h"
#include "hub_cmd.h"
#include "hub_to_hub.h"
#include "multi_private_chat.h"
#include "multi_public_chat.h"
#include "emb_perl.h"
#ifdef WITH_PYTHON
#include "emb_python.h"
#endif
#include "hub_cluster.h"

#include "color_debug.h"


/*******************************************************/
/* send the first queued outgoing string to the socket */
/*******************************************************/
/* output: =0: ok     */
/*         !=0: error */
/**********************/
static int process_outgoing_cnx_data(CNX_ENTRY *ptr)
{
	GString_ref *str;
	int ret=0;
#ifndef MSG_NOSIGNAL
	void (*old_handler)(int);
#endif

#if 0
	printf("process_outgoing.\n");
#endif
	str=take_first_string_of_glist_outgoing(ptr,&(ptr->outgoing_commands));


	if(str==NULL)
		return 0;			/* having nothing to send is not a fatal error */

	/* send the message in one part, it should not be broken into serveral parts */
#ifdef __COLOR_DEBUG_H__
#if 0
	printf(COLORED1 "send : %s%d"COLONONE"\n",str->data->str, ptr->sock_fd);
#endif
#endif

#ifdef MSG_NOSIGNAL
	if (send(ptr->sock_fd,str->data->str,str->data->len,/* MSG_WAITALL| */ MSG_NOSIGNAL)!=str->data->len)
		ret=-1;
#else
	old_handler = signal (SIGPIPE, SIG_IGN);
	if (send(ptr->sock_fd,str->data->str,str->data->len, 0)!=str->data->len)
		ret=-1;
	signal (SIGPIPE, old_handler);
#endif
	ptr->out_data += str->data->len;
	g_string_ref_free(str);
#if 0
	printf("process_outgoing done.\n");
#endif
	return ret;
}

/*********************************************************************/
/* get data from the socket, merge them with the partially received  */
/* one, split strings along | and put them in incoming_commands list */
/*********************************************************************/
/* output: =0: ok     */
/*         !=0: error */
/**********************/
static int process_incoming_cnx_data(CNX_ENTRY *ptr)
{
	char buf[8192];
#ifndef MSG_NOSIGNAL
	void (*old_handler)(int);
#endif
	int ret;
	char *t;

#if 0
	printf("process_incoming.\n");
#endif

#ifdef MSG_NOSIGNAL
	ret=recv(ptr->sock_fd, buf,sizeof(buf)-1,MSG_NOSIGNAL);
#else
	old_handler = signal (SIGPIPE, SIG_IGN);
	ret=recv(ptr->sock_fd, buf,sizeof(buf)-1,0);
	signal (SIGPIPE, old_handler);
#endif

	if(ret==-1)
	{
		return -1;
	}

	if(ret==0)
	{
		/* printf("CNX CLOSE\n"); */
		return -2;		/* connection closed */
	}
	if(ret>0)
		buf[ret]='\0';
	ptr->in_data += ret;

	if(ptr->in_partial==NULL)
		ptr->in_partial=g_string_new(buf);
	else
		ptr->in_partial=g_string_append(ptr->in_partial,buf);

#ifdef __COLOR_DEBUG_H__
#if 0
	{
		int i;
		for(i=0;i<ret;i++)
		{
			if(!isprint(buf[i]))
				buf[i]='.';
		}
		printf(COLOGREEN1 "received: %s %d"COLONONE"\n",buf, ptr->sock_fd);
	}
#endif
#endif
	if (ptr->in_partial) /* flood protection */
	{
		if (ptr->in_partial->len > 8192)
		{
			if (!(ptr->is_op & BOT_PRIV))
			{
				unsigned int hip;
		
				hip=ntohl(ptr->ip.s_addr);
				printf("loflood: <%s> %hhu.%hhu.%hhu.%hhu %lu %u/%u:%u/%u\n", (ptr->user_nick!=NULL)?ptr->user_nick->str:"",
							(unsigned char)(hip>>24)&0xff,
							(unsigned char)(hip>>16)&0xff,
							(unsigned char)(hip>>8)&0xff,
							(unsigned char)hip&0xff,
							gl_cur_time - ptr->start_time,
							ptr->out_data,
							ptr->in_data,
							ptr->out_cnx_count,
							ptr->in_cnx_count
							);
				shutdown(ptr->sock_fd, 0);
			}
		}
	}
	/* split incoming string along | */
	t=strchr(ptr->in_partial->str,'|');
	while(t!=NULL)
	{
		GString *nw;

		/* future new content of in_partial */
		/* = old content w/o the first string ending w/ a | */
		if (ptr->in_partial->len == (t-ptr->in_partial->str+1)) /*add in 0.1.4.b2*/
			nw =NULL;
		else
			nw=g_string_new(t+1);

		/* new string to add to incoming_commands */
		/* = first string ending w/ a | */
		ptr->in_partial=g_string_truncate(ptr->in_partial,(t+1)-ptr->in_partial->str);

		append_new_cmd_to_glist_incoming(ptr,&(ptr->incoming_commands),ptr->in_partial->str);
		g_string_free(ptr->in_partial,TRUE);

		ptr->in_partial=nw;
		if (nw != NULL) /*add in 0.1.4.b2*/
			t=strchr(ptr->in_partial->str,'|');
		else
			t = NULL;
	}

#if 0
	printf("process_incoming done.\n");
#endif
	return 0;
}

/*********************************************/
/* this function performs input/output tasks */
/*************************************************************************/
/* input: ptr= entry to manage                                           */
/*        no_read_wait= ==0: wait until something comes from remote side */
/*                      ==1: only send local data and return immediately */
/*        no_send_retry= Only have a real meaning when no_read_wait==1   */
/*                       ==0: the command won't return until all queued  */
/*                            commands to send have not been sent.       */
/*                       ==1: make 1 try to send the queued commands and */
/*                            return even if data was not sent.          */
/*************************************************************************/
/* output: =0: ok     */
/*         !=0: error */
/**********************/
int cnx_io_processing(CNX_ENTRY *ptr, int no_read_wait, int no_send_retry)
{
	fd_set rd,wd;
	struct timeval tv;
	int n;
	int ln;
	const GString_ref *str;
	int ret=0;

#if 0
	printf("cnx_io_process.\n");
#endif
	while(1)
	{
		n=-1;
		FD_ZERO(&rd);
		FD_ZERO(&wd);

		/* wanna read ? */
		if(no_read_wait==0)
		{
			FD_SET(ptr->sock_fd,&rd);
			n=max(n,ptr->sock_fd);
		}

		/* something to send ? */
		str=peek_first_string_of_glist_outgoing(ptr,&(ptr->outgoing_commands));
		if(str!=NULL)
		{
			FD_SET(ptr->sock_fd,&wd);
			n=max(n,ptr->sock_fd);
		}
		else
		{
			if(peek_first_string_of_glist_incoming(ptr,&(ptr->incoming_commands))!=NULL)
			{
				/* nothing to send and something ready to process, we can leave the loop */
				break;
			}
		}

		if(n==-1)		/* no more things to do, just leave */
			break;

		tv.tv_sec=0;
		tv.tv_usec=100000;

		ln=select(n+1,&rd,&wd,NULL,&tv);
		if(ln>0)
		{
			if(FD_ISSET(ptr->sock_fd,&wd))
			{
				if(process_outgoing_cnx_data(ptr)!=0)
				{
					ret=-1;
					break;
				}
			}

			if((no_read_wait==0)&&(FD_ISSET(ptr->sock_fd,&rd)))
			{
				if(process_incoming_cnx_data(ptr)!=0)
				{
					ret=-2;
					break;
				}
			}
		}
		else
		{
			if(no_send_retry==1)
			{
				ret=-3;
				break;
			}
		}
	}
#if 0
	printf("cnx_io_process done.\n");
#endif
	return ret;
}

/*****************************************/
/* check if too many users are connected */
/*************************************************/
/* output: 0=no, 1=yes                           */
/* if ==1, this function has handled redirection */
/*************************************************/
static int too_many_users(CNX_ENTRY *ptr)
{
	int cur_nb;
	int max_nb;

	cur_nb=get_local_nb_connected();
	
	G_LOCK(conf_file);
	if(!e_db_int_get(conf_file,"MAXUSER",&max_nb))
		max_nb=100;
	G_UNLOCK(conf_file);

	if(cur_nb>max_nb)
	{
		int v;
		/* too many user connected */
		G_LOCK(conf_file);
		if(!e_db_int_get(conf_file,"REDIR_FULL",&v))
			v=0;
		G_UNLOCK(conf_file);

		if(v==0)
		{
			/* display an error message */
			disp_msg:
			send_const_str_to(ptr,"<Hub-Security> This hub is currently full!|");
			send_const_str_to(ptr,"$HubIsFull|");
		}
		else
		{
			char *rd;

			G_LOCK(conf_file);
			rd=e_db_str_get(conf_file,"REDIR_ADDR");
			G_UNLOCK(conf_file);
			if(rd==NULL)
				goto disp_msg;
			if(strlen(rd)==0)
			{
				free(rd);
				goto disp_msg;
			}

			/* send a forcemove with the destination address */
			{
				GString *rd_msg;
				rd_msg=g_string_new("");

				/* first the message to display */
				g_string_sprintf(rd_msg,"<Hub-Security> This hub is currently full! You are being redirected to %s.|",rd);
				send_const_str_to(ptr,rd_msg->str);

				/* and then the redirection command */
				g_string_sprintf(rd_msg,"$ForceMove %s|",rd);
				send_const_str_to(ptr,rd_msg->str);

				g_string_free(rd_msg,TRUE);
			}
			free(rd);
		}
		cnx_io_processing(ptr,1,0);		/* just send stored data */
		return 1;
	}
	else
		return 0;
}

typedef struct
{
	const char *cmd;
	size_t cmd_len;
	int (*fnc)(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param);
		/* function to call. This function MUST free str when it is not useful anymore */
		/* the function must return 0 if the connection must remain alive (no error or */
		/* recoverable error) and a value !=0 if there is a fatal error) */
	char *xtra_param;
} KNOWN_CMD;

/*************************/
/* process $Search query */
/*************************/
static int search_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	GString *org_str;
	GString_ref *org_str2 = 0;
	char *t;
	char *u;
	char *v;
	int ret=1;

	ptr->sr_count++;
	if (((gl_cur_time - ptr->last_hcall) < gl_sr_interval) && ((ptr->is_op & BOT_PRIV)==0))
	{
		send_const_str_to(ptr,"<Kernel> search dropped. Wait 10sec between 2 searchs|");
		g_string_free(str,TRUE);
		return (0);
	}
	ptr->last_hcall = gl_cur_time;
	org_str=g_string_new(str->str);

	/* a $Search query has the following format: */
	/* $Search Hub:nick a?b?c?d?eeeee| where */
	/* a is F if size doesn't matter, else T */
	/* b is F if size is "at least", else T (at most) */
	/* c is the size in byte */
	/* d is data type: 1=any,2=audio,3=compressed,4=document,5=exe,6=picture,7=videos,8=folder */
	/* and eeee is the pattern to find */

	t=str->str+strlen("$Search ");
	SKIP_SPACE(t);
	u=strchr(t,':');
	if(u==NULL)
		goto leave;
	v=strchr(t,' ');
	if(v==NULL)
		goto leave;
	if(u>v)
		goto leave;
	t=v;				/* skip the return address (Hub:nick or hostip:port) */
	SKIP_SPACE(t);

	if((*t!='F')&&(*t!='T'))	/* size matter ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	if((*t!='F')&&(*t!='T'))	/* size at least or at most ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted size */
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted type */
	if(t==NULL)
		goto leave;
	t++;

	if(strlen(t)<2)			/* nothing to search  (or just the trailing | ) ? */
		goto leave;

	/* the query is valid, send it */
	
	org_str2 = conv_gstring_to_gstring_ref(org_str);
	queue_to_everyone(org_str2);
	ret=0;

	send_evt_to_script("search",ptr->user_nick->str,1,org_str->str);
#ifdef WITH_PYTHON
	send_evt_to_pscript("search",ptr->user_nick->str,1,org_str->str);
#endif
	leave:
	if (org_str2 != 0)
		g_string_ref_free(org_str2);
	else
		g_string_free(org_str, TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/*************************/
/* process $UniSearch query */
/*************************/
static int unisearch_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	GString *org_str;
	char *t;
	char *u;
	char *v;
	char *destnick;
	int ret=1;
	char *beginning_std_search;

	org_str=g_string_new(str->str);

	/* a $UniSearch query has the following format: */
	/* $UniSearch destnick ip:port a?b?c?d?eeeee| where */
	/* a is F if size doesn't matter, else T */
	/* b is F if size is "at least", else T (at most) */
	/* c is the size in byte */
	/* d is data type: 1=any,2=audio,3=compressed,4=document,5=exe,6=picture,7=videos,8=folder */
	/* and eeee is the pattern to find */

	t=str->str+strlen("$UniSearch ");
	SKIP_SPACE(t);

	destnick=t;
	t=strchr(t,' ');
	if(t==NULL)
		goto leave;
	*t++='\0';
	SKIP_SPACE(t);

	beginning_std_search=t;

	u=strchr(t,':');
	if(u==NULL)
		goto leave;
	v=strchr(t,' ');
	if(v==NULL)
		goto leave;
	if(u>v)
		goto leave;
	t=v;				/* skip the return address (Hub:nick or hostip:port) */
	SKIP_SPACE(t);

	if((*t!='F')&&(*t!='T'))	/* size matter ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	if((*t!='F')&&(*t!='T'))	/* size at least or at most ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted size */
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted type */
	if(t==NULL)
		goto leave;
	t++;

	if(strlen(t)<2)			/* nothing to search  (or just the trailing | ) ? */
		goto leave;

	/* the query is valid, send it */
	{
		GString *rebuild_str;
		rebuild_str=g_string_new("");
		g_string_sprintf(rebuild_str,"$Search %s",beginning_std_search);
		send_str_to_named_user(destnick, rebuild_str->str);
		g_string_free(rebuild_str,TRUE);
	}
	ret=0;

	send_evt_to_script("unisearch",ptr->user_nick->str,1,org_str->str);
#ifdef WITH_PYTHON
	send_evt_to_pscript("unisearch",ptr->user_nick->str,1,org_str->str);
#endif
	leave:
	g_string_free(org_str,TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/*************************/
/* process $MultiSearch query */
/*************************/
static int multisearch_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	GString *org_str;
	char *t;
	char *u;
	char *v;
	int ret=1;

	org_str=g_string_new(str->str);

	/* a $Search query has the following format: */
	/* $Search Hub:nick a?b?c?d?eeeee| where */
	/* a is F if size doesn't matter, else T */
	/* b is F if size is "at least", else T (at most) */
	/* c is the size in byte */
	/* d is data type: 1=any,2=audio,3=compressed,4=document,5=exe,6=picture,7=videos,8=folder */
	/* and eeee is the pattern to find */

	t=str->str+strlen("$Search ");
	SKIP_SPACE(t);
	u=strchr(t,':');
	if(u==NULL)
		goto leave;
	v=strchr(t,' ');
	if(v==NULL)
		goto leave;
	if(u>v)
		goto leave;
	t=v;				/* skip the return address (Hub:nick or hostip:port) */
	SKIP_SPACE(t);

	if((*t!='F')&&(*t!='T'))	/* size matter ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	if((*t!='F')&&(*t!='T'))	/* size at least or at most ? T or F */
		goto leave;
	t=strchr(t,'?');
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted size */
	if(t==NULL)
		goto leave;
	t++;

	t=strchr(t,'?');			/* skip the wanted type */
	if(t==NULL)
		goto leave;
	t++;

	if(strlen(t)<2)			/* nothing to search  (or just the trailing | ) ? */
		goto leave;

	/* the query is valid, send it */
	send_string_to_all_rhubs(org_str);
	ret=0;

	send_evt_to_script("multisearch",ptr->user_nick->str,1,org_str->str);
#ifdef WITH_PYTHON
	send_evt_to_pscript("multisearch",ptr->user_nick->str,1,org_str->str);
#endif
	leave:
	g_string_free(org_str,TRUE);
	g_string_free(str,TRUE);
	return ret;
}


/**************************/
/* process $Version query */
/**************************/
int version_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	str=g_string_erase(str,0,strlen(cmd));
	if(str->len>0)
	{
		str=g_string_truncate(str,str->len-1);		/* remove trailing | */
	}
	
	ptr->client_version=g_string_assign(ptr->client_version,str->str);
	g_string_free(str,TRUE);
	return 0;
}

static const char *lst_cnx_type[]=
                           {
                              "56Kbps",
                              "33.6Kbps",
                              "28.8Kbps",
                              "Satellite",      /* according to DC, passive mode is required here */
                              "ISDN",
                              "DSL",
                              "Cable",
                              "LAN(T1)",
                              "LAN(T3)",
                              NULL
                           };

static const char *fnd_static_cnx_type(char *type)
{
	int i;

	i=0;
	while(lst_cnx_type[i]!=NULL)
	{
		if(!strcmp(lst_cnx_type[i],type))
			return lst_cnx_type[i];
		i++;
	}

	return "";
}

/*************************/
/* process $MyINFO query */
/*********************************************************************/
/* string format:  $MyINFO $ALL nickname aaa$ $xxxf$bbb$yyy$         */
/*       aaa is user description                                     */
/*       xxx is connection type (ex: Cable)                          */
/*       f is a 1 byte flag. Default value: \x01                     */
/*             if not behind firewall, bit 1 must be set, else clear */
/*       bbb is e-mail (empty string)                                */
/*       yyy is size of shared data in bytes                         */
/*********************************************************************/
int myinfo_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *t;
	char *fnd_nick;
	char *fnd_desc;
	char *fnd_cnx_type;
	unsigned char fnd_flag;
	char *fnd_email;
	char *fnd_size;
	int ret=0;
	int have_changed=0;

	t=str->str;
	if(strncmp(t,"$MyINFO $ALL ",strlen("$MyINFO $ALL ")))
	{
		have_err:
		ret=1;
		goto eofunc;
	}

	t+=strlen("$MyINFO $ALL ");
	SKIP_SPACE(t);

	/* extract the nickname */
	fnd_nick=t;
	t=strchr(fnd_nick,' ');
	if(t==NULL)
		goto have_err;
	*t++='\0';
	SKIP_SPACE(t);

	if (ptr->server_id)
	{
		ptr=get_in_cnx_hash_table_by_nick(cnx_lst, fnd_nick);
		if (ptr == 0)
		{
			printf ("MEGAMERDE\n");
			return (0);
		}
	}

	/* extract the description */
	fnd_desc=t;
	t=strchr(fnd_desc,'$');
	if(t==NULL)
		goto have_err;
	*t++='\0';

	if(strncmp(t," $",2))
		goto have_err;
	t+=2;

	/* extract the connection type */
	fnd_cnx_type=t;
	t=strchr(fnd_cnx_type,'$');
	if(t==NULL)
		goto have_err;
	*t++='\0';

	/* extract the email */
	fnd_email=t;
	t=strchr(fnd_email,'$');
	if(t==NULL)
		goto have_err;
	*t++='\0';

	/* extract the shared size */
	fnd_size=t;
	t=strchr(fnd_size,'$');
	if(t==NULL)
		goto have_err;
	*t++='\0';

	/* well, nearly all fields have been extracted */
	/* only fnd_flag still is merged with fnd_cnx_type */
	if(strlen(fnd_cnx_type)>1)
	{
		fnd_flag=fnd_cnx_type[strlen(fnd_cnx_type)-1];
		fnd_cnx_type[strlen(fnd_cnx_type)-1]='\0';
	}
	else
		goto have_err;

	/* ok, it's time to check parameters */
	if(strcmp(fnd_nick,ptr->user_nick->str))
		goto have_err;			/* nickname doesn't match -> abort */

	/* update information only when needed */
	{
		const char *t;
		unsigned long long new_share_size;

		t=fnd_static_cnx_type(fnd_cnx_type);
		if(ptr->user_cnx_type!=t)					/* t and user_cnx_type point on const static string, that's why we can avoid a strcmp */
		{
			ptr->user_cnx_type=fnd_static_cnx_type(fnd_cnx_type);
			have_changed=1;
		}

		sscanf(fnd_size,"%Lu",&(new_share_size));
		if(new_share_size!=ptr->shared_size)
		{
			ptr->shared_size=new_share_size;
			have_changed=1;
		}

		if(strcmp(ptr->user_mail->str,fnd_email))
		{
			ptr->user_mail=g_string_assign(ptr->user_mail,fnd_email);
			have_changed=1;
		}

		if(strcmp(ptr->user_description->str,fnd_desc))
		{
			ptr->user_description=g_string_assign(ptr->user_description,fnd_desc);
			have_changed=1;
		}

		if(ptr->user_flag!=(((unsigned int)fnd_flag)&255))
		{
			ptr->user_flag=((unsigned int)fnd_flag)&255;
			have_changed=1;
		}
	}
	
	if(have_changed)
	{
		GString *nw;
		GString_ref *nw2;
#ifdef __CYGWIN__
		char		long_int[48];
#endif

		nw=g_string_new("");
#ifdef __CYGWIN__
		sprintf(long_int, "%Lu", ptr->shared_size);
		g_string_sprintf(nw,"$MyINFO $ALL %s %s$ $%s%c$%s$%s$|",
#else
		g_string_sprintf(nw,"$MyINFO $ALL %s %s$ $%s%c$%s$%Lu$|",
#endif
									ptr->user_nick->str,
									ptr->user_description->str,
									ptr->user_cnx_type,
									(char)ptr->user_flag,
									ptr->user_mail->str,
#ifdef __CYGWIN__
				 					long_int
#else
				 					ptr->shared_size
#endif
				 					);
		nw2 = conv_gstring_to_gstring_ref(nw);
		if (ptr->server_id)
			queue_to_everyone(nw2);
		else
			queue_to_local_everyone(nw2);
		send_evt_to_script("myinfo",ptr->user_nick->str,1,nw->str);
#ifdef WITH_PYTHON
		send_evt_to_pscript("myinfo",ptr->user_nick->str,1,nw->str);
#endif
		g_string_ref_free(nw2);
	}
	eofunc:
	g_string_free(str,TRUE);
	return ret;
}

/******************************/
/* process $GetNickList query */
/******************************/
static int getnicklist_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	GString *ulst;
	GString *oplst;
	int i;

	/* a fake user named "Hub-Security" is created instead DC admin cannot send command in chat */
	/* build a $NickList with the list of all users */
	ulst=g_string_new("$NickList ");

	/* and the $OpList with the list of all operators */
	oplst=g_string_new("$OpList ");

	/* only operator and master can see the hub-security "fake" user */
	if(ptr->is_op & OP_PRIV)
	{
		ulst=g_string_append(ulst,"Hub-Security$$");
		oplst=g_string_append(oplst,"Hub-Security$$");
	}

	for(i=0;i<HASH_SIZE;i++) /* for getnicklist_process */
	{
		CNX_ENTRY *ent;
		for(ent=cnx_lst->entry[i]; ent; ent=ent->next)
		{
			if((ent==NULL)||(ent->user_nick==NULL)||(ent->user_nick->len==0))
				continue;

			ulst=g_string_append(ulst,ent->user_nick->str);
			ulst=g_string_append(ulst,"$$");

			if(ent->is_op & OP_PRIV)
			{
				oplst=g_string_append(oplst,ent->user_nick->str);
				oplst=g_string_append(oplst,"$$");
			}
		}
	}

	append_public_chat_name(ulst, ptr);

	ulst=g_string_append_c(ulst,'|');
	send_const_str_to(ptr,ulst->str);
	g_string_free(ulst,TRUE);

	oplst=g_string_append_c(oplst,'|');
	send_const_str_to(ptr,oplst->str);
	g_string_free(oplst,TRUE);

	g_string_free(str,TRUE);

	send_evt_to_script("getnicklist",ptr->user_nick->str,0);
#ifdef WITH_PYTHON
	send_evt_to_pscript("getnicklist",ptr->user_nick->str,0);
#endif
	return 0;
}


static int allow_small_event(CNX_ENTRY *ptr)
{
	if ((ptr->is_op & BOT_PRIV) == 0)
	{
		ptr->msg_count++;
		if (ptr->last_msg == gl_cur_time)
		{
			if (ptr->msg_count >= gl_silent_limit)
			{
				if (ptr->msg_count > gl_kick_limit)
				{
					unsigned int hip;

					hip=ntohl(ptr->ip.s_addr);
					printf("ccflood: <%s> %hhu.%hhu.%hhu.%hhu %lu %u/%u:%u/%u\n", (ptr->user_nick!=NULL)?ptr->user_nick->str:"",
			 						(unsigned char)(hip>>24)&0xff,
			 						(unsigned char)(hip>>16)&0xff,
			 						(unsigned char)(hip>>8)&0xff,
			 						(unsigned char)hip&0xff,
			 						gl_cur_time - ptr->start_time,
			 						ptr->out_data,
			 						ptr->in_data,
			 						ptr->out_cnx_count,
			 						ptr->in_cnx_count
			 						);
					shutdown(ptr->sock_fd, 0);
				}
				return 0;
			}
		}
		else
		{
			long int diff;
	
			diff = gl_cur_time - ptr->last_msg;
			if (ptr->msg_count > diff)
				ptr->msg_count -= diff;
			else
				ptr->msg_count = 0;
			ptr->last_msg = gl_cur_time;
		}
#if 0
		else if ((gl_cur_time - ptr->last_msg) == 1)
		{
			ptr->msg_count--;
		}
		else if ((gl_cur_time - ptr->last_msg) == 2)
		{
			if (ptr->msg_count >= 2)
				ptr->msg_count -= 2;
			else
				ptr->msg_count = 0;
		}
		else
		{
			ptr->msg_count = 0;
		}
#endif
	}
	return(1);
}


static int force_send_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char	*t;

	if ((ptr->is_op & BOT_PRIV) == 0)
		return (1);
	t=str->str+strlen("$ForceSend ");
	queue_const_str_to_everyone(t);
	return(0);
}

static int force_send_to_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char	*t;
	char	*n;
	CNX_ENTRY *dest;
	
	if ((ptr->is_op & BOT_PRIV) == 0)
		return (1);
	n=t=str->str+strlen("$ForceSendTo ");

	while(*t && (*t != ' '))
		t++;

	if (!((t[0]) && (t[1])))
		return (1);

	t[0] = '\0';
	dest=find_cnx_by_nickname(n);
	t[0] = ' ';

	if (dest == NULL)
		return (1);
	t++;
	send_const_str_to(dest,t);
	return(0);
}

/*******************/
/* process < query */
/*******************/
static int global_chat_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	int ret=1;

	if ((gl_moderate_mode) && (ptr->is_op == 0))
	{
		send_const_str_to(ptr,"<Kernel> the central chat is in moderate mode|");
		goto global_chat_process_fin;
	}
	if (allow_small_event(ptr) == 0)
		goto global_chat_process_fin;
	if(ptr->user_nick!=NULL)
	{
		if (!strncmp(str->str+1, ptr->user_nick->str, ptr->user_nick->len) || (ptr->is_op & BOT_PRIV))
		{
			if(!strncmp(&(str->str[1+ptr->user_nick->len]),"> ",2) || (ptr->is_op & BOT_PRIV))
			{
				GString_ref *str2;
				str2 = conv_gstring_to_gstring_ref(str);
				/* the string starts with "<nickname> " */
				if (ptr->is_op & BOT_PRIV)
				{
					if (strchr(str->str, '>') == 0)
						goto global_chat_process_fin;
					queue_to_everyone(str2);
				}
				else
				{
					if (str->str[1+ptr->user_nick->len+2]!='+')
						queue_to_everyone(str2);
					else
						process_user_hub_command(ptr,&str->str[1+ptr->user_nick->len+2]);
				}
				send_evt_to_script("globalchat",ptr->user_nick->str,1,str->str);
#ifdef WITH_PYTHON
				send_evt_to_pscript("globalchat",ptr->user_nick->str,1,str->str);
#endif
				ret=0;
				g_string_ref_free(str2);
				return ret;
			}
		}
	}

 	global_chat_process_fin:
	g_string_free(str,TRUE);
	return ret;
}

/*********************/
/* process $To query */
/*********************/
int private_chat_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *dest_nick;
	char *source_nick;
	char *t;
	GString *org_cmd;
	int ret=1;

	org_cmd=g_string_new(str->str);
	
	if(ptr->user_nick==NULL)
		goto leave;


	if (ptr->is_op == 0)
	{
		if (allow_small_event(ptr) == 0)
			goto leave;
	}

	/* extract the dest nickname and the source nickname */
	t=str->str+strlen("$To: ");
	SKIP_SPACE(t);
	dest_nick=t;
	t=strchr(dest_nick,' ');
	if(t==NULL)
		goto leave;

	*t++='\0';
	if(strncmp(t,"From: ",strlen("From: ")))
		goto leave;

	t+=strlen("From: ");
	SKIP_SPACE(t);
	source_nick=t;
	t=strchr(source_nick,' ');
	if(t==NULL)
		goto leave;

	*t++='\0';
	if(*t!='$')
		goto leave;

	t++;
	if(strcmp(source_nick,ptr->user_nick->str) && ((ptr->is_op & BOT_PRIV) == 0))
		goto leave;			/* the source nickname does not match with connexion nickname */
	
	if(!strcasecmp(dest_nick,"Hub-Security"))
	{
		ret=process_hub_command(ptr,t);
	}
	else
	{
		char *skip_beginning;

		skip_beginning=strchr(t,' ');
		SKIP_SPACE(skip_beginning);
		{
			int idx;

			if ((idx = find_pubchat_idx_by_name(dest_nick)) != -1)
			{
				send_to_users_of_multi_pubchat_except2(idx, t, ptr);
			}
		}

		if(skip_beginning[0]!='+')
			send_str_to_named_user(dest_nick,org_cmd->str);
		else
			process_user_hub_command(ptr,skip_beginning);
		send_evt_to_script("privchat",ptr->user_nick->str,2,dest_nick,t);
#ifdef WITH_PYTHON
		send_evt_to_pscript("privchat",ptr->user_nick->str,2,dest_nick,t);
#endif
		ret=0;
	}

	leave:
	g_string_free(org_cmd,TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/******************************************************************/
/* process $GetINFO query and send the $MyINFO of the wanted user */
/*********************************************************************/
/* string format:  $MyINFO $ALL nickname aaa$ $xxxf$bbb$yyy$         */
/*       aaa is user description                                     */
/*       xxx is connection type (ex: Cable)                          */
/*       f is a 1 byte flag. Default value: \x01                     */
/*             if not behind firewall, bit 1 must be set, else clear */
/*       bbb is e-mail (empty string)                                */
/*       yyy is size of shared data in bytes                         */
/*********************************************************************/
	static int get_info_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *wanted_nick;
	char *mynick;
	char *t;
	int ret=1;

	t=str->str+strlen("$GetINFO ");
	SKIP_SPACE(t);
	wanted_nick=t;
	t=strchr(wanted_nick,' ');
	if(t==NULL)
		goto leave;
	*t++='\0';
	SKIP_SPACE(t);
	mynick=t;
	
	t=strchr(mynick,'|');
	if(t!=NULL)
	{
		CNX_ENTRY *wnt;
		*t='\0';

		if(strcmp(mynick,ptr->user_nick->str))
			goto leave;

		wnt=find_cnx_by_nickname(wanted_nick);
		if(wnt!=NULL)
		{
			GString *output;
			GString_ref *output2;
#ifdef __CYGWIN__
			char		long_int[48];
#endif

			output=g_string_new("");
#ifdef __CYGWIN__
			sprintf(long_int, "%Lu", wnt->shared_size);
			g_string_sprintf(output,"$MyINFO $ALL %s %s$ $%s%c$%s$%s$|",
					 wnt->user_nick->str,
					 wnt->user_description->str,
					 wnt->user_cnx_type,
					 (char)wnt->user_flag,
					 wnt->user_mail->str,
					 long_int);
#else
			g_string_sprintf(output,"$MyINFO $ALL %s %s$ $%s%c$%s$%Lu$|",
										wnt->user_nick->str,
										wnt->user_description->str,
										wnt->user_cnx_type,
										(char)wnt->user_flag,
										wnt->user_mail->str,
										wnt->shared_size);
#endif
			output2 = conv_gstring_to_gstring_ref(output);
			send_str_to(ptr,output2);
			g_string_ref_free(output2);
			ret=0;
		}
	}
	
	leave:
	g_string_free(str,TRUE);
	return ret;
}

/***************************/
/* process $RevConnectToMe */
/***************************/
int revconnecttome_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *wanted_nick;
	char *mynick;
	char *t;
	GString *org_str;
	int ret=1;

	ptr->out_cnx_count++;
	org_str=g_string_new(str->str);
	t=str->str+strlen("$RevConnectToMe ");
	SKIP_SPACE(t);
	mynick=t;
	t=strchr(mynick,' ');
	if(t==NULL)
		goto leave;
	*t++='\0';
	SKIP_SPACE(t);
	wanted_nick=t;
	
	t=strchr(wanted_nick,'|');
	if(t!=NULL)
	{
		*t='\0';

		if(strcmp(mynick,ptr->user_nick->str))
			goto leave;


		{
			CNX_ENTRY *ent;
		
			ent = send_str_to_named_user(wanted_nick,org_str->str);
			if (ent)
			{
				if (ent != ptr)
					ent->in_cnx_count++;
				else
				{
					/* That hides something ....*/
					printf("clientbug: %s\n", ent->user_nick->str);
				}
			}
		}

		send_evt_to_script("revconnect",ptr->user_nick->str,2,wanted_nick,org_str->str);
#ifdef WITH_PYTHON
		send_evt_to_pscript("revconnect",ptr->user_nick->str,2,wanted_nick,org_str->str);
#endif
		ret=0;
	}
	
	leave:
	g_string_free(org_str,TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/************************/
/* process $ConnectToMe */
/************************/
int connecttome_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *wanted_nick;
	char *myip;
	char *t;
	GString *org_str;
	int ret=1;

	ptr->out_cnx_count++;
	org_str=g_string_new(str->str);
	t=str->str+strlen("$ConnectToMe ");
	SKIP_SPACE(t);
	wanted_nick=t;
	t=strchr(wanted_nick,' ');
	if(t==NULL)
		goto leave;
	*t++='\0';
	SKIP_SPACE(t);
	myip=t;
	
	t=strchr(myip,'|');
	if(t!=NULL)
	{
		*t='\0';
		{
			CNX_ENTRY *ent;
		
			ent = send_str_to_named_user(wanted_nick,org_str->str);
			if (ent)
			{
				if (ent != ptr)
					ent->in_cnx_count++;
				else
				{
					/* That hides something ....*/
					printf("clientbug: %s\n", ent->user_nick->str);
				}
			}
		}

		send_evt_to_script("connecttome",ptr->user_nick->str,2,wanted_nick,org_str->str);
#ifdef WITH_PYTHON
		send_evt_to_pscript("connecttome",ptr->user_nick->str,2,wanted_nick,org_str->str);
#endif
		ret=0;
	}
	
	leave:
	g_string_free(org_str,TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/*****************/
/* process $Kick */
/**********************************/
/* string format:  $Kick nickname */
/**********************************/
static int kick_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *wanted_nick;
	char *t;
	int ret=1;

	t=str->str+strlen("$Kick ");
	SKIP_SPACE(t);
	wanted_nick=t;
	t=strchr(wanted_nick,'|');
	if(t!=NULL)
	{
		GString *query;
		*t='\0';

		/* translate the command into hub command */
		query=g_string_new("");
		g_string_sprintf(query,"<%s> -kick \"%s\" ...|",ptr->user_nick->str,wanted_nick);
		process_hub_command(ptr,query->str);
		g_string_free(query,TRUE);
		ret=0;
	}
	
	g_string_free(str,TRUE);
	return ret;
}

/***************/
/* process $SR */
/**************************************************************************************************/
/* string format:  $SR nickname filename\005filesize slot/ratio\005hubname (hub addr)\005destnick */
/**************************************************************************************************/
static int sr_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *t;
	int ret=1;
	GString *org;
	

	/* keep a copy of the incoming string. This string will be sent to the dest nick */
	org=g_string_new(str->str);

	t=str->str+strlen("$SR ");
	SKIP_SPACE(t);

#define SKIP_TO_NEXT(t,sep1,sep2,want)											\
										while((*t)&&(*t!=sep1)&&(*t!=sep2))\
											t++;\
										if(*t!=want)\
											goto abrt;\
											t++;

	/* we are on the source nickname */
	SKIP_TO_NEXT(t,' ',5,' ')

	/* we are on the filename */
	t=strchr(t,5);
	if(t==NULL)
		goto abrt;
	t++;

	/* we are on the filesize */
	SKIP_TO_NEXT(t,' ',5,' ')

	/* we are on the slot ratio */
	SKIP_TO_NEXT(t,' ',5,5)

	/* we are on the hubname */
	t=strchr(t,5);
	if(t==NULL)
		goto abrt;

	/* modify the $SR to create the one sent to the client */
	org=g_string_truncate(org,t-str->str);
	org=g_string_append_c(org,'|');

	str=g_string_erase(str,0,t-str->str+1);	/* keep everything after the separator */
	t=str->str;
	t=strchr(t,'|');
	if(t!=NULL)
	{
		*t='\0';
		send_str_to_named_user(str->str,org->str);
		ret=0;
	}
	else
	{
		abrt:
		/*
		all directory SR are invalid It's a DC hub bug.
		printf("invalid $SR result\n");
		*/
	}
	
	g_string_free(org,TRUE);
	g_string_free(str,TRUE);
	return ret;
}

/************************/
/* process $OpForceMove */
/************************************************************************/
/* string format:  $OpForceMove $Who:nick$Where:redirectip$Msg:message| */
/************************************************************************/
/* Added by [Tele2]Shoggot*/
/**************************/
static int redir_process(const char *cmd, CNX_ENTRY *ptr, GString *str, char *xtra_param)
{
	char *t;
	char *temp;
	int ret=1;
	char nick[21];
	char ip[122];
	char message[11];
	GString *query2;
	query2=g_string_new("");
	t=str->str;

	sscanf(t, "$OpForceMove $Who:%20[^$]$Where:%121[^$]$Msg:%10[^|]|", nick, ip, message);
	
	if(message[0] == '\0')
	{
		printf("Erroneous op_force_move command received\n");
		g_string_free(str,TRUE);
		g_string_free(query2,TRUE);
		return ret;
	}

	if((temp = strstr(t, "$Msg:")) == NULL)
	{
		printf("Erroneous op_force_move command received\n");
		g_string_free(str,TRUE);
		g_string_free(query2,TRUE);
		return ret;
	}

	g_string_sprintf(query2,"<%s> -redir \"%s\" \"%s\"|",ptr->user_nick->str,nick,ip);
	process_hub_command(ptr,query2->str);
	g_string_free(query2,TRUE);
	ret=0;

	g_string_free(str,TRUE);
	return ret;
}


static KNOWN_CMD known_cmd[]=	{
											{"$Version ",			sizeof("$Version ")-1,			version_process,			NULL},
											{"$Search ",			sizeof("$Search ")-1,			search_process,			NULL},
											{"$UniSearch ",		sizeof("$UniSearch ")-1,		unisearch_process,		NULL},
											{"$MultiSearch ",		sizeof("$MultiSearch ")-1,		multisearch_process,		NULL},
											{"$MyINFO ",			sizeof("$MyINFO ")-1,			myinfo_process,			NULL},
											{"$GetNickList",		sizeof("$GetNickList")-1,		getnicklist_process,		NULL},
											{"<",						sizeof("<")-1,						global_chat_process,		NULL},
											{"$To: ",				sizeof("$To: ")-1,				private_chat_process,	NULL},
											{"$GetINFO ",			sizeof("$GetINFO ")-1,			get_info_process,			NULL},
											{"$RevConnectToMe ",	sizeof("$RevConnectToMe ")-1,	revconnecttome_process,	NULL},
											{"$ConnectToMe ",		sizeof("$ConnectToMe ")-1,		connecttome_process,		NULL},
											{"$Kick ",				sizeof("$Kick ")-1,				kick_process,				NULL},
											{"$SR ",					sizeof("$SR ")-1,					sr_process,					NULL},
											{"$OpForceMove ",		sizeof("$OpForceMove ")-1,		redir_process,				NULL},
											{"$ForceSend ",		sizeof("$ForceSend ")-1,		force_send_process,		NULL},
											{"$ForceSendTo ",		sizeof("$ForceSendTo ")-1,		force_send_to_process,	NULL},
											{NULL,0,NULL,NULL},
};

/************************************************************************/
/* this function takes the first incoming command and choose what to do */
/************************************************************************/
/* return value: <0: fatal error; ==0: ok; >0: non fatal error */
/***************************************************************/
static int process_one_command(CNX_ENTRY *ptr)
{
	GString *inp;
	int i;
	int fnd=0;
	int ret=0;

	inp=take_first_string_of_glist_incoming(ptr,&(ptr->incoming_commands));
	if(inp==NULL)
		return ret;		/* this case should never appears but who knows ... */

	if (ptr->connected)
	{
		if (!strncmp(inp->str, "brd:", strlen("brd:")))
			return (parse_brd_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "new:", strlen("new:")))
			return (parse_new_entry_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "ne2:", strlen("ne2:")))
			return (parse_new_entry_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "rem:", strlen("rem:")))
			return (parse_rem_entry_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "pvm:", strlen("pvm:")))
			return(parse_pvm_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "kik:", strlen("kik:")))
			return(parse_kik_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "ban:", strlen("ban:")))
			return(parse_ban_from_hub(inp, ptr));
		else if (!strncmp(inp->str, "hub:", strlen("hub:"))) /* CNX type 2 asked */
			return(parse_hub_from_hub(inp, ptr));
		return (0);
	}
	
	i=0;
	while(known_cmd[i].cmd!=NULL)
	{
		if(!strncmp(inp->str,known_cmd[i].cmd,known_cmd[i].cmd_len))
		{
			fnd=1;
#ifdef DEBUG
			printf("command from %s: %s\n",ptr->user_nick->str,inp->str);
#endif
			ret=(known_cmd[i].fnc)(known_cmd[i].cmd,ptr,inp,known_cmd[i].xtra_param);
#ifdef DEBUG
			printf("eocmmnd from %s: %d\n",ptr->user_nick->str,ret);
#endif
			break;
		}
		i++;
	}

	if(!fnd)
	{
		send_evt_to_script("unknowncmd",ptr->user_nick->str,1,inp->str);
#ifdef WITH_PYTHON
		send_evt_to_pscript("unknowncmd",ptr->user_nick->str,1,inp->str);
#endif
#if 0
		printf("unknown command: %s\n",inp->str);
#endif
		g_string_free(inp,TRUE);
	}
	return ret;
}


/*
  return the maximum fd
 */

int	cnxs_fnc_process_mask(CNX_ENTRY *ptr, int n, fd_set *rd, fd_set *wd)
{
	const GString_ref *str;

				/* wanna read ? */
	FD_SET(ptr->sock_fd,rd);
				n=max(n,ptr->sock_fd);

				/* something to send ? */
	str=peek_first_string_of_glist_outgoing(ptr,&(ptr->outgoing_commands));
	if (str!=NULL)
	{
		FD_SET(ptr->sock_fd,wd);
		n=max(n,ptr->sock_fd);
	}
	return (n);
}

/*
RETURN 0 IF OK
RETURN 1 if the entry had been removed
*/

int	cnxs_fnc_process_cmd(CNX_ENTRY *ptr)
{
	if(peek_first_string_of_glist_incoming(ptr,&(ptr->incoming_commands))!=NULL)
	{
		if(process_one_command(ptr)<0)
		{
			GString *msg;
			GString_ref *msg2;

			msg=g_string_new("");
			g_string_sprintf(msg,"$Quit %s|",ptr->user_nick->str);
			msg2 = conv_gstring_to_gstring_ref(msg);
			queue_to_everyone(msg2);
			g_string_ref_free(msg2);
			send_evt_to_script("quit",ptr->user_nick->str,0);
#ifdef WITH_PYTHON
			send_evt_to_pscript("quit",ptr->user_nick->str,0);
#endif

		/* printf("cnx over1.\n"); */
			{
				unsigned int hip;

				hip=ntohl(ptr->ip.s_addr);
				printf("cnxover1: <%s> \t%hhu.%hhu.%hhu.%hhu %lu %d %u/%u:%u/%u\n", (ptr->user_nick!=NULL)?ptr->user_nick->str:"",
			 					(unsigned char)(hip>>24)&0xff,
			 					(unsigned char)(hip>>16)&0xff,
			 					(unsigned char)(hip>>8)&0xff,
			 					(unsigned char)hip&0xff,
			 					gl_cur_time - ptr->start_time,
			 					ptr->sr_count,
			 					ptr->out_data,
			 					ptr->in_data,
			 					ptr->out_cnx_count,
			 					ptr->in_cnx_count
							);
			}
			delete_cnx_entry(ptr);
			return (1);
		}
	}
	return (0);
}

int	lost_hub(CNX_ENTRY* ptr);

/*
RETURN 0 IF OK
RETURN 1 if the entry had been removed
*/
int	cnxs_fnc_process_io(fd_set *rd, fd_set *wd, CNX_ENTRY *ptr)
{
	if(FD_ISSET(ptr->sock_fd,wd))
	{
		FD_CLR(ptr->sock_fd,wd);
		if (process_outgoing_cnx_data(ptr)!=0)
			goto this_user_leaves;
	}
	if(FD_ISSET(ptr->sock_fd,rd))
	{
		FD_CLR(ptr->sock_fd, rd);
		if (process_incoming_cnx_data(ptr)!=0)
		{
			/* after an error, the user may be removed from the cnx list */
			this_user_leaves:
			if (ptr->log_level<4)
			{
				unsigned int hip;

				hip=ntohl(ptr->ip.s_addr);
				printf("cnxover3: <%s> \t%hhu.%hhu.%hhu.%hhu %lu %d %u/%u:%u/%u\n", 
				 				(ptr->user_nick!=NULL)?ptr->user_nick->str:
				 				(ptr->temp_nick!=NULL)?ptr->temp_nick:"",
				 				(unsigned char)(hip>>24)&0xff,
				 				(unsigned char)(hip>>16)&0xff,
				 				(unsigned char)(hip>>8)&0xff,
				 				(unsigned char)hip&0xff,
				 				gl_cur_time - ptr->start_time,
				 				ptr->sr_count,
				 				ptr->out_data,
				 				ptr->in_data,
				 				ptr->out_cnx_count,
				 				ptr->in_cnx_count
				 				);
				remove_from_all_cnx_table(cnx_lst, ptr);
				return(1); /* goto onemoretime */
			}
			else
			{
				GString *msg;
				GString_ref *msg2;
	
				if (ptr->connected)
				{
					printf("cnxover5\n");

					lost_hub(ptr);
					return(0);
				}
				msg=g_string_new("");
				g_string_sprintf(msg,"$Quit %s|",ptr->user_nick->str);
				msg2 = conv_gstring_to_gstring_ref(msg);
				queue_to_everyone(msg2);
				g_string_ref_free(msg2);
				send_evt_to_script("quit",ptr->user_nick->str,0);
#ifdef WITH_PYTHON
				send_evt_to_pscript("quit",ptr->user_nick->str,0);
#endif
				{
					unsigned int hip;

					hip=ntohl(ptr->ip.s_addr);
					printf("cnxover1: <%s> \t%hhu.%hhu.%hhu.%hhu %lu %d %u/%u:%u/%u\n", (ptr->user_nick!=NULL)?ptr->user_nick->str:"",
					 			(unsigned char)(hip>>24)&0xff,
					 			(unsigned char)(hip>>16)&0xff,
					 			(unsigned char)(hip>>8)&0xff,
					 			(unsigned char)hip&0xff,
					 			gl_cur_time - ptr->start_time,
					 			ptr->sr_count,
					 			ptr->out_data,
					 			ptr->in_data,
					 			ptr->out_cnx_count,
					 			ptr->in_cnx_count
					 			);
				}
				delete_cnx_entry(ptr);
				return(1); /* goto onemoretime */
			}
		}
	}
	return (0);
}

/******************************************************/
/* check if the client reply to $Lock is a valid $Key */
/******************************************************/
/* output: 0 = data structure had change */
/*****************************************/

int is_ip(char *str)
{
	unsigned int ip[4];
	if (sscanf(str, "%u.%u.%u.%u", ip + 0, ip + 1, ip + 2, ip + 3) == 4)
		return(1);
	return (0);
}

/******************************************/
/* main part of the connection management */
/******************************************/
#define DEBUG
int	new_cnx(CNX_ENTRY *ptr)
{
	GString *inp;
	CNX_ENTRY *gost_ent = 0;
	/* check the reply. It must be a valid $Key value */

	inp=take_first_string_of_glist_incoming(ptr,&(ptr->incoming_commands));
	if(inp==NULL)
		return 1;		/* nothing = error */

	if (ptr->log_level == -5) /* RCV KEY */
	{
		gchar *user_str;
		GString *key;
		char	*t;

		/* Compute RCV key */
		if ((inp->len<10) || strncmp(inp->str,"$Lock ",6))
		{
			printf("invalid Key received from HUB 1\n");
			inter_hub_failure:
			g_string_free(inp,TRUE);
			route_table->table[ptr->server_id].user_flag = 0;
			remove_from_all_cnx_table(cnx_lst, ptr);
			return(0);
		}
		inp=g_string_erase(inp,0,6);
		t=strchr(inp->str,' ');
		if (t==NULL)
		{
			printf("invalid Key received from HUB 1 esp not found\n");
			goto inter_hub_failure;
		}
		if((t-inp->str)<3)
		{
			printf("invalid Key received from HUB : Lock too small\n");
			goto inter_hub_failure;
		}
		inp=g_string_truncate(inp,t-inp->str);
		key=compute_access_key(inp);
		key=g_string_prepend(key,"$Key ");
		key=g_string_append_c(key,'|');
		send_const_str_to(ptr,key->str);
		g_string_free(key,TRUE);
		g_string_free(inp,TRUE);
		if (fusion_user == 0)
		{
			printf("fusion_user is empty interHUB CNX failed\n");
			goto inter_hub_failure;
		}
		user_str=g_strconcat("$ValidateNick ", fusion_user, "|",NULL);
		send_const_str_to(ptr,user_str);
		g_free(user_str);
		ptr->log_level = -6;
		return(-1);
	}
	if (ptr->log_level == -6) /* RCV KEY */
	{
		gchar *passwd_str;

		if (inp->str[0] == '<') /* ignore msg RCV	"<Hub-Security> You are using a password protected nickname, please enter the password.|"; */
		{
			g_string_free(inp,TRUE);
			return(1);
		}
		if (strncmp(inp->str,"$GetPass", sizeof("$GetPass")-1) == 0) /* RCV	"$GetPass|"; */
		{
			passwd_str=g_strconcat("$MyPass ", fusion_pass, "|",NULL);
			send_const_str_to(ptr,passwd_str);
			g_free(passwd_str);
			g_string_free(inp,TRUE);
			ptr->log_level = -7;
			return(1);
		}
		goto inter_hub_failure;
	}
	if (ptr->log_level == -7) /* RCV LoginIn */
	{
		int		i;
		CNX_ENTRY		*ent;

		if (strncmp(inp->str, "$LogedIn", sizeof("$LogedIn")-1) == 0) /* RCV	"$LogedIn ....|"; */
		{
			if (ptr->user_flag == 1) /* send if TYPE I */
			{
				char	*url;

				G_LOCK(conf_file);
				url=e_db_str_get(conf_file,"HUBADDR");
				G_UNLOCK(conf_file);
				if (url)
				{
					gchar *annonce_str;
					char	str_port[10];

					sprintf(str_port, ":%hu", hub_port);
					annonce_str=g_strconcat("$AnnonceMe ", url , str_port, "|",NULL);
					/* printf("level -7, send : %s\n", annonce_str); */
					send_const_str_to(ptr,annonce_str);
					g_free(annonce_str);
					free(url);
				}
				else
				{
					printf("unable to localize myselft \"HUBADDR\" is empty\n");
					goto inter_hub_failure;
				}
			}
			else if (ptr->user_flag == 2) /* send if type II */
			{
				printf("level -7 TYPE II\n");
				send_const_str_to(ptr, "$Watashi |"); 
			}
			else
			{
				printf("error interhub cnx != 1 and !=2\n");
				goto inter_hub_failure;
			}
			/* send the whole local user list */		/* send userlist locaux LOCAL -> NEW (4) */
			for(i=0;i<HASH_SIZE;i++) 
			{
				for(ent=cnx_lst->entry[i];ent;ent=ent->next)
				{
					if((ent->sock_fd!=-1)&&(ent->server_id==0))	
						send_new_entry_to_hub(ent, ptr, 0); /* vue2 */
				}
			}
	
			ptr->log_level = 5;
			ptr->handshake_done=1;		/* give the control to the main thread */	
			g_string_free(inp,TRUE);
			inject_wait_table(cnx_lst);
			ptr->log_level = 5;
			return(0);
		}
	}
	/* ------------------------- */
	/* stage 1: get a valid $Key */
	/* ------------------------- */
	if (ptr->log_level == 0)
	{
		int lret;

		lret=verify_key(ptr->created_lock,inp);
		g_string_free(inp,TRUE);
		if (lret!=0)
		{
			/* printf("key chk lailure\n"); */
			goto login_failure;
		}
		if (ptr->created_lock)
			free(ptr->created_lock);
		ptr->created_lock=0;
		ptr->log_level = 1;
		return (1);
	}

	/* ----------------------------------	*/
	/* stage 2: get a valid $ValidateNick	*/
	/* ----------------------------------	*/
	if (ptr->log_level == 1)
	{
		char *etn;

		if (strncmp(inp->str,"$ValidateNick ",sizeof("$ValidateNick ")-1))
		{
			g_string_free(inp,TRUE);
			printf("$ValidateNick expected but (%s) found\n", inp->str);
			goto login_failure;
		}
		/* theoritically, ' ' and '$' are not allowed in string */
		ptr->temp_nick=strdup(inp->str+strlen("$ValidateNick "));
		if ((strchr(ptr->temp_nick, ' '))!=NULL)
		{
			invalid_char:
			printf("' ' or '$' or '' in nick.\n");
			g_string_free(inp,TRUE); /* error 6 */
			goto login_failure;
		}
		if((strchr(ptr->temp_nick, '$'))!=NULL)
			goto invalid_char;
		if((strchr(ptr->temp_nick, ''))!=NULL) /* the  character is neaded by the regexp lock */
			goto invalid_char;
		if((strchr(ptr->temp_nick, '\"'))!=NULL) /* username with " can not be kick with the Hub-security*/
			goto invalid_char;
		if (strlen(ptr->temp_nick) > 47) /* dc protocol do not allow username longer than 47 char */
			goto invalid_char;

		etn=strchr(ptr->temp_nick,'|');
		if (etn == NULL)
		{
			printf("SEGV imminent\n");
		}
		*etn='\0';		/* very dirty to do this on a GString :) */
		/* check if the nick is not already used on the hub */
		if ((fusion_user) && ((strcasecmp(ptr->temp_nick, fusion_user)==0)))
		{
			ptr->tmp_op_type=4;
			if (fusion_pass)
				ptr->user_pass=strdup(fusion_pass);
			printf("new HUB in queue\n");
		}
		else if ((is_a_mchat_name(ptr->temp_nick)) ||
				 (is_a_pubchat_name(ptr->temp_nick)) ||
				 (!strcasecmp("Hub-Security",ptr->temp_nick)) || 
				 (!strcasecmp("Hub-Mass-Message",ptr->temp_nick)) ||
				 ((gost_ent = find_cnx_by_nickname(ptr->temp_nick))) ||
				 is_ip(ptr->temp_nick))
		{
			/* nick not valid (already in use) (or is a multi-user chat name)*/
			/* printf ("Nick %s already in use\n",ptr->temp_nick); */

			if ((gost_ent) && (gost_ent->ip.s_addr == ptr->ip.s_addr))
			{
				delete_cnx_entry(gost_ent);
				goto gost_rem;
			}
			{
				gchar *validedenide;
				validedenide=g_strconcat("$ValidateDenide ",ptr->temp_nick,"|",NULL);
				send_const_str_to(ptr,validedenide);
				g_free(validedenide);
			}
			g_string_free(inp,TRUE);	/* don't forget to free inp */
			return(1);
		}


		gost_rem:
		g_string_free(inp,TRUE);	/* don't forget to free inp */
		ptr->log_level=2;
	}

	/* ---------------------------------- */
	/* stage 3: passwd check needed? srv	*/
	/* ---------------------------------- */
	if (ptr->log_level == 2) /* replay from the hub */
	{
		int only_reg;

		/* check if this user is a registered user (password protected) */
		if (ptr->tmp_op_type != 4)
			ptr->user_pass=get_user_password(ptr->temp_nick,&ptr->tmp_op_type);

		/* check if too many users on the hub */
		if((ptr->user_pass==NULL)||(ptr->tmp_op_type==0)) /* only normal users (registered or not) are limited in number */
		{
			if(too_many_users(ptr))
			{
				cnx_io_processing(ptr,1,0);		/* send the reply before leaving */
				goto login_failure; /* error 11 */
			}
		}

		/* check if hub is public or not */
		G_LOCK(conf_file);
		if(!e_db_int_get(conf_file,"ONLY_REG",&only_reg))
			only_reg=0;
		G_UNLOCK(conf_file);

		if (only_reg != 0 && ptr->user_pass == NULL)
		{
			/* only reg users can connect */
			printf("Unregistered user: %s (ONLY_REG HUB)\n",ptr->temp_nick);
			send_const_str_to(ptr,"<Hub-Security> This Hub is for registered users only.|");
			cnx_io_processing(ptr,1,0);		/* send the reply before leaving */
			goto login_failure; /* error 10 */
		}

		if (ptr->user_pass != NULL)
		{	/* this is a password protected user */
			if (strlen(ptr->user_pass) > 0 || (ptr->tmp_op_type > 0))
			{ /* OP and MASTER must enter password */

				/* send password query */
				send_const_str_to(ptr,"<Hub-Security> You are using a password protected nickname, please enter the password.|");
				send_const_str_to(ptr,"$GetPass|");
				ptr->log_level = 3;
				return (1);
			}
		}
		else /* std user */
		{
			ptr->log_level = 4;
		}
	} /* end log_level == 2 */

	/* ---------------------------------- */
	/* stage 4: passwd needed cli				 */
	/* ---------------------------------- */
	if (ptr->log_level == 3) /* replay from the hub*/
	{
		char *etn;

		if (strncmp(inp->str,"$MyPass ",sizeof("$MyPass ")-1))
		{
			printf("no $MyPass returned on $GetPass (%s obtained)\n",inp->str);
			g_string_free(inp,TRUE);
			goto login_failure; /* error 9 */
		}
		/* remove trash before the password */
		inp=g_string_erase(inp,0,sizeof("$MyPass ")-1);
		etn=strchr(inp->str,'|');
		*etn='\0';		/* very dirty to do this on a GString :) */
			
		if(strcmp(inp->str, ptr->user_pass))
		{
			printf("Invalid password (%s) for %s. Wanted: %s\n",
		 				inp->str,ptr->temp_nick,ptr->user_pass);
			g_string_free(inp,TRUE);				
			send_const_str_to(ptr,"$BadPass|");
			cnx_io_processing(ptr,1,0);		/* send the reply before leaving */
			goto login_failure; /* error 10 */
		}

		{
			unsigned int hip;

			hip=ntohl(ptr->ip.s_addr);
			printf("valid_pass %s for %s %hhu.%hhu.%hhu.%hhu\n", 
				 		ptr->user_pass,
				 		ptr->temp_nick,
				 		(unsigned char)(hip>>24)&0xff,
				 		(unsigned char)(hip>>16)&0xff,
				 		(unsigned char)(hip>>8) &0xff,
				 		(unsigned char) hip		 &0xff);
		}

		/* freeing unuseful string */
		g_string_free(inp,TRUE);
		free(ptr->user_pass);
		ptr->user_pass=0;
			
		/* ok, the password is valid */
		{
			GString *okpass;
			GString_ref *okpass2;
	
			okpass=g_string_new("$LogedIn ");
			okpass=g_string_append(okpass,ptr->temp_nick);
			okpass=g_string_append_c(okpass,'|');
			okpass2 = conv_gstring_to_gstring_ref(okpass);
			send_str_to(ptr,okpass2);
			g_string_ref_free(okpass2);
			ptr->is_op=ptr->tmp_op_type;
			ptr->log_level = 4;
			if (ptr->is_op & HUB_PRIV) /* external hub */ /* CODE EXPERIMENTAL */
			{
				printf ("Hub login\n");
				ptr->log_level = -1;
				ptr->server_id = allocate_new_hub(ptr);
				printf("HUB added : %d/%d\n", ptr->server_id, route_table->max_server_id);
				return(1);
			}
		}
	} /* end log_level == 3 */
	if (ptr->log_level == -1) /* replay from the hub*/
	{
		if (strncmp(inp->str,"$AnnonceMe ",sizeof("$AnnonceMe ")-1)==0)
		{		
			int		i;
			CNX_ENTRY	*ent;
			char		*url;

			printf(">>>>>>>>>>level -1 TYPE I\n");
			url = inp->str + sizeof("$AnnonceMe ")-1;
			if (route_table->server_count > 1)		/* ask to the other hub to connect to the new hub LOCAL -> HUBS (2) */
			{
				int i;
				GString *paked_str;
				GString_ref *paked_str2;
			
				printf("enter FOLLOW new hub\n");
				paked_str = g_string_new("");
				g_string_sprintf(paked_str,"hub:%s|", url);
				paked_str2=conv_gstring_to_gstring_ref(paked_str);
				for (i = 1; i <= route_table->max_server_id; i++)
				{
					if (route_table->table[i].connected)
						send_str_to((route_table->table + i),paked_str2);
				}
				g_string_ref_free(paked_str2);
			}
			printf("annonceMe done\n");
			for(i=0;i<HASH_SIZE;i++) /* send the whole local user list */		/* send userlist locaux LOCAL -> NEW (4) */
			{
				for(ent=cnx_lst->entry[i];ent;ent=ent->next)
				{
					if((ent->sock_fd!=-1)&&(ent->server_id==0))
						send_new_entry_to_hub(ent, ptr, 1); /* vue2 */
				}
			}
			inject_wait_table(cnx_lst);
			ptr->log_level = 5;
		}
		else if (strncmp(inp->str,"$Watashi ",sizeof("$Watashi ")-1) == 0)
		{
			int		i;
			CNX_ENTRY	*ent;

			printf("level -1 TYPE II\n");

			for(i=0;i<HASH_SIZE;i++) /* send the whole local user list */		// send userlist locaux LOCAL -> NEW (4)
			{
				for(ent=cnx_lst->entry[i];ent;ent=ent->next)
				{
					if((ent->sock_fd!=-1)&&(ent->server_id==0))
						send_new_entry_to_hub(ent, ptr, 1);
				}
			}
			printf("fusion 3-----------------------------\n");
		}
		else
		{
			g_string_free(inp,TRUE);
			route_table->table[ptr->server_id].user_flag = 0;		
			remove_from_all_cnx_table(cnx_lst, ptr);
			return(0);
		}
		g_string_free(inp,TRUE);
		ptr->handshake_done=1;		/* give the control to the main thread */
		inject_wait_table(cnx_lst);
		ptr->log_level = 5;
		return(0);
	}

	if (ptr->log_level == -2)
	{
		if (!strncmp(inp->str, "new:", sizeof("new:")-1))
		{
			parse_new_entry_from_hub(inp, ptr);			/* return a list	LOCAL <- NEW RCV (5) */
			return(1);
		}
		if (!strncmp(inp->str, "end:", sizeof("end:")-1))
		{
			/* inject hub */
			ptr->log_level = 5;
			return (0);
		}
	}
	 
	if (ptr->log_level == 4) /* replay from the hub*/
	{
		/* as soon as we have assigned the user nickname, the user is connected */

		ptr->user_nick=g_string_assign(ptr->user_nick, ptr->temp_nick);
		free(ptr->temp_nick);
		ptr->temp_nick = 0;

		/* send this user the hub name */
		send_const_str_to(ptr,hubname->str);
		send_const_str_to(ptr,"<Hub-Security> This hub is running version 1.0.25 of the NeoModus Direct Connect Hub Software.|");
			
		send_const_str_to(ptr,"<welcome> This hub is running " PACKAGE " " VERSION". build on "__DATE__" at "__TIME__".|");

		/* we must also send this to all users of the hub */
		{
			GString *msg;
			GString_ref *msg2;

			msg=g_string_new("");
			/* if (ptr->server_id) g_string_sprintf(msg,"%s$$Hello %s|",ptr->user_nick->str,ptr->user_nick->str); else  FIXED ? */
			g_string_sprintf(msg,"$Hello %s|",ptr->user_nick->str);
			msg2 = conv_gstring_to_gstring_ref(msg);
			queue_to_everyone(msg2);
			g_string_ref_free(msg2);
		}
		if (ptr->is_op && OP_PRIV)
		{
			GString *msg;
			GString_ref *msg2;
			GString *oplst;
			GString_ref *oplst2;
			int i;

			/* Send the $OpList with the list of all operators */
			oplst=g_string_new("$OpList ");
			for(i=0;i<HASH_SIZE;i++) /* FIXME LOW SPEED create OP LIST */
			{
				CNX_ENTRY *ent;
				for(ent=cnx_lst->entry[i]; ent; ent=ent->next)
				{
					if((ent==NULL)||(ent->user_nick==NULL)||(ent->user_nick->len==0))
						continue;
					if (ent->is_op & OP_PRIV)
					{
						oplst=g_string_append(oplst,ent->user_nick->str);
						oplst=g_string_append(oplst,"$$");
					}
				}
			}

			oplst=g_string_append_c(oplst,'|');
			oplst2 = conv_gstring_to_gstring_ref(oplst);
			queue_to_everyone(oplst2);
			g_string_ref_free(oplst2);
		
			msg=g_string_new("");
			g_string_sprintf(msg,
				 "$To: %s From: Hub-Security $Welcome %s\r\n"
				 "\r\n"
				 "Using this private chat, you can enter special commands:\r\n"
				 "Use -help for the command list.\r\n|",
				 ptr->user_nick->str,
				 (ptr->is_op & MASTER_PRIV)?"Master of the hub":"operator");
			msg2 = conv_gstring_to_gstring_ref(msg);
			send_str_to(ptr,msg2);
			g_string_ref_free(msg2);
		}
			
		/* Send welcome message, reads from database */
		{
			char *rd;
			GString *msg;
			GString_ref *msg2;

			G_LOCK(conf_file);
			rd=e_db_str_get(conf_file,"WELCOME");
			G_UNLOCK(conf_file);
			if((rd!=NULL)&&(strlen(rd)!=0))
			{
				msg=g_string_new("");
				g_string_sprintf(msg,"$To: %s From: Hub $%s|",ptr->user_nick->str,rd);
				msg2 = conv_gstring_to_gstring_ref(msg);
				send_str_to(ptr,msg2);
				free(rd);
				g_string_ref_free(msg2);
			}
		}

		ptr->handshake_done=1;		/* give the control to the main thread */
		ptr->last_hcall = gl_cur_time;
		inject_wait_table(cnx_lst);
		return(0);
	} /* end log_level == 4 */

 login_failure:
	{
		unsigned int hip;

		hip=ntohl(ptr->ip.s_addr);
		printf("loginfailure%d <%s> %hhu.%hhu.%hhu.%hhu\n",
				 ptr->log_level,
				 (ptr->temp_nick == NULL)?"":ptr->temp_nick,
				 (unsigned char)(hip>>24)&0xff,
				 (unsigned char)(hip>>16)&0xff,
				 (unsigned char)(hip>>8) &0xff,
				 (unsigned char) hip		 &0xff);
	}
	remove_from_all_cnx_table(cnx_lst, ptr);
	return(0);
}
